"""
The `~certbot_dns_schlundtech.dns_schlundtech` plugin automates the process of
completing a ``dns-01`` challenge (`~acme.challenges.DNS01`) by creating, and
subsequently removing, TXT records using the SchlundTech XML Gateway API.

Named Arguments
---------------

===========================================  =================================
``--dns-schlundtech-credentials``            SchlundTech credentials_ INI
                                             file.
                                             (Required)
``--dns-schlundtech-propagation-seconds``    The number of seconds to wait for
                                             DNS to propagate before asking
                                             the ACME server to verify the DNS
                                             record.
                                             (Default: 60)
===========================================  =================================


Credentials
-----------
Use of this plugin requires a configuration file containing SchlundTech XML
Gateway API credentials:
* ``user``
* ``password``
* ``context``
* ``token``

.. code-block:: ini
   :name: credentials.ini
   :caption: Example credentials config file:
   # Credentials
   dns_schlundtech_user = 54321
   dns_schlundtech_password = PASSWORD
   dns_schlundtech_context = 10
   dns_schlundtech_token = SECRETTOKEN

The path to this file can be provided interactively or using the
`--dns-schlundtech-credentials` command-line argument. Certbot records the
path to this file for use during renewal, but does not store the file's
contents.

.. caution::
   You should protect these credentials. Users who can read this file can use
   these credentials to issue some types of API calls on your behalf, limited
   by the permissions assigned to the account. Users who can cause Certbot to
   run using these credentials can complete a ``dns-01`` challenge to acquire
   new certificates or revoke existing certificates for domains these
   credentials are authorized to manage.


Examples
--------
.. code-block:: bash
   :caption: To acquire a certificate for ``example.com``
   certbot certonly \\
     --server https://acme-v02.api.letsencrypt.org/directory \\
     -a dns-schlundtech \\
     --dns-schlundtech-credentials \\
       ~/.secrets/certbot/schlundtech.ini \\
     -d example.com
.. code-block:: bash
   :caption: To acquire a single certificate for both ``example.com`` and
             ``www.example.com``
   certbot certonly \\
     --server https://acme-v02.api.letsencrypt.org/directory \\
     -a dns-schlundtech \\
     --dns-schlundtech-credentials \\
       ~/.secrets/certbot/schlundtech.ini \\
     -d example.com \\
     -d www.example.com
.. code-block:: bash
   :caption: To acquire a certificate for ``example.com``, waiting 60 seconds
             for DNS propagation
   certbot certonly \\
     --server https://acme-v02.api.letsencrypt.org/directory \\
     -a dns-schlundtech \\
     --dns-schlundtech-credentials \\
       ~/.secrets/certbot/schlundtech.ini \\
     --dns-schlundtech-propagation-seconds 60 \\
     -d example.com
"""



"""Tests for certbot_dns_schlundtech.dns_schlundtech."""
import certbot.compat.os as os
import unittest

import mock
from certbot import errors
from certbot.plugins import dns_test_common
from certbot.plugins.dns_test_common import DOMAIN
from certbot.tests import util as test_util

patch_display_util = test_util.patch_display_util

USER = 'user'
PASSWORD = 'password'
CONTEXT = 10
TOKEN = 'token'
TTL = 10


class AuthenticatorTest(test_util.TempDirTestCase, dns_test_common.BaseAuthenticatorTest):

    def setUp(self):
        from certbot_dns_schlundtech.dns_schlundtech import Authenticator

        super(AuthenticatorTest, self).setUp()

        path = os.path.join(self.tempdir, 'file.ini')
        dns_test_common.write({
            "dns_schlundtech_user": USER,
            "dns_schlundtech_password": PASSWORD,
            "dns_schlundtech_context": CONTEXT,
            "dns_schlundtech_token": TOKEN,
        }, path)

        self.config = mock.MagicMock(dns_schlundtech_credentials=path, dns_schlundtech_propagation_seconds=0)
        self.auth = Authenticator(self.config, "dns-schlundtech")

        self.mock_client = mock.MagicMock()
        # _get_gateway_client | pylint: disable=protected-access
        self.auth._get_gateway_client = mock.MagicMock(return_value=self.mock_client)

    @patch_display_util()
    def test_perform(self, unused_mock_get_utility):
        self.auth.perform([self.achall])

        expected = [mock.call.add_txt_record(DOMAIN, '_acme-challenge.'+DOMAIN, mock.ANY)]
        self.assertEqual(expected, self.mock_client.mock_calls)

    @patch_display_util()
    def test_cleanup(self, unused_mock_get_utility):
        # _attempt_cleanup | pylint: disable=protected-access
        self.auth._attempt_cleanup = True
        self.auth.cleanup([self.achall])

        expected = [mock.call.del_txt_record(DOMAIN, '_acme-challenge.'+DOMAIN, mock.ANY)]
        self.assertEqual(expected, self.mock_client.mock_calls)


class SchlundtechGatewayClientTest(unittest.TestCase):
    record_name = "_acme-challenge"
    record_fqdn = record_name + "." + DOMAIN
    record_content = "bar"
    system_ns = "ns1." + DOMAIN

    def setUp(self):
        from certbot_dns_schlundtech.dns_schlundtech import _SchlundtechGatewayClient
        self.gateway_client = _SchlundtechGatewayClient(USER, PASSWORD, CONTEXT, TOKEN, TTL)
        self._mock_call({})  # Safety mock

    def test_auth(self):
        self.assertDictEqual(
            {
                'user': USER,
                'password': PASSWORD,
                'context': CONTEXT,
                'token': TOKEN
            },
            self.gateway_client._auth()
        )

    def test_zone_info(self):
        self._mock_zone_info('success', {'name': DOMAIN})
        zone = self.gateway_client._zone_info(DOMAIN, self.record_fqdn)
        self.assertDictEqual(zone, {'name': DOMAIN})

    def test_zone_info_domain_missing(self):
        self._mock_zone_info('error')
        self.assertRaises(errors.PluginError, self.gateway_client._zone_info, DOMAIN, self.record_fqdn)

    def test_add_txt_record(self):
        self._mock_zone_info('success', {'name': DOMAIN, 'system_ns': self.system_ns, 'soa': {'level': '1'}}, False)
        self._mock_call({'status': {'type': 'success'}})
        self.gateway_client.add_txt_record(DOMAIN, self.record_name, self.record_content)
        self.gateway_client._call.assert_called_with({
            'code': '0202001',
            'zone': {
                'name': DOMAIN,
                'system_ns': self.system_ns
            },
            'default': {
                'rr_add': {
                    'name': self.record_name,
                    'type': 'TXT',
                    'value': self.record_content,
                    'ttl': TTL
                },
                'soa': {'level': '1'}
            }
        })

    def test_add_txt_record_domain_missing(self):
        self._mock_zone_info('error')
        self.assertRaises(errors.PluginError,
                          self.gateway_client.add_txt_record, DOMAIN, self.record_name, self.record_content)

    def test_add_txt_record_already_exists(self):
        self._mock_zone_info('success',
                             {
                                 'name': DOMAIN, 'system_ns': self.system_ns, 'soa': {'level': '1'},
                                 'rr': [{'name': self.record_name, 'value': self.record_name}]
                             },
                             False)
        self._mock_call({'status': {'type': 'success'}})
        self.gateway_client.add_txt_record(DOMAIN, self.record_name, self.record_content)
        self.gateway_client._call.assert_called_with({
            'code': '0202001',
            'zone': {
                'name': DOMAIN,
                'system_ns': self.system_ns
            },
            'default': {
                'rr_add': {
                    'name': self.record_name,
                    'type': 'TXT',
                    'value': self.record_content,
                    'ttl': TTL
                },
                'soa': {'level': '1'}
            }
        })

    def test_add_txt_record_error(self):
        self._mock_zone_info('success', {'name': DOMAIN, 'system_ns': self.system_ns, 'soa': {'level': '1'}}, False)
        self._mock_call({'status': {'type': 'error'}})
        self.assertRaises(errors.PluginError,
                          self.gateway_client.add_txt_record, DOMAIN, self.record_name, self.record_content)

    def test_del_txt_record(self):
        self._mock_zone_info('success', {'name': DOMAIN, 'system_ns': self.system_ns, 'soa': {'level': '1'}}, False)
        self._mock_call({'status': {'type': 'success'}})
        self.gateway_client.del_txt_record(DOMAIN, self.record_name, self.record_content)
        self.gateway_client._call.assert_called_with({
            'code': '0202001',
            'zone': {
                'name': DOMAIN,
                'system_ns': self.system_ns
            },
            'default': {
                'rr_rem': {
                    'name': self.record_name,
                    'type': 'TXT',
                    'value': self.record_content,
                    'ttl': TTL
                }
            }
        })

    def test_del_txt_record_domain_missing(self):
        self._mock_zone_info('error')
        self.assertRaises(errors.PluginError,
                          self.gateway_client.del_txt_record, DOMAIN, self.record_name, self.record_content)

    def test_del_txt_record_error(self):
        self._mock_zone_info('success', {'name': DOMAIN, 'system_ns': self.system_ns, 'soa': {'level': '1'}}, False)
        self._mock_call({'status': {'type': 'error'}})
        self.assertRaises(errors.PluginError,
                          self.gateway_client.del_txt_record, DOMAIN, self.record_name, self.record_content)

    def _mock_zone_info(self, status, zone=None, mock_call=True):
        if mock_call:
            self._mock_call({'status': {'type': status}, 'data': {'zone': zone}})
        else:
            # _zone_info | pylint: disable=protected-access
            self.gateway_client._zone_info = mock.MagicMock(return_value=zone)

    def _mock_call(self, return_value):
        # _call | pylint: disable=protected-access
        self.gateway_client._call = mock.MagicMock(return_value=return_value)


class XmlTest(unittest.TestCase):
    tag = 'test'
    data = '<' + tag + '><a>1</a><a>2</a><b>hello</b></' + tag + '>'
    obj = {'a': ['1', '2'], 'b': 'hello'}
    encoding = 'UTF-8'

    def setUp(self):
        from certbot_dns_schlundtech.dns_schlundtech import _XML
        self.xml = _XML()

    def test_fromstring(self):
        r = self.xml.fromstring(self.data)
        self.assertDictEqual(self.obj, r)

    def test_tostring(self):
        r = self.xml.tostring(self.tag, self.obj).decode(self.encoding)
        self.assertEqual(self.data, r)


if __name__ == '__main__':
    unittest.main()  # pragma: no cover



"""DNS Authenticator for the SchlundTech XML Gateway."""
import logging
import xml.etree.ElementTree as Et
try:
    from urllib.request import Request, urlopen
    from urllib.error import HTTPError, URLError
except ImportError:
    from urllib2 import Request, urlopen, HTTPError, URLError

import zope.interface
from certbot import errors
from certbot import interfaces
from certbot.plugins import dns_common

logger = logging.getLogger(__name__)

GATEWAY_URL = 'https://gateway.schlundtech.de'


@zope.interface.implementer(interfaces.IAuthenticator)
@zope.interface.provider(interfaces.IPluginFactory)
class Authenticator(dns_common.DNSAuthenticator):
    """DNS Authenticator for the SchlundTech XML Gateway.

    This Authenticator uses the SchlundTech XML Gateway API to fulfill a dns-01 challenge.
    """
    description = 'Obtain certificates using a DNS TXT record (if you are using SchlundTech for DNS).'
    ttl = 60

    def __init__(self, *args, **kwargs):
        super(Authenticator, self).__init__(*args, **kwargs)
        self.credentials = None

    @classmethod
    def add_parser_arguments(cls, add):  # pylint: disable=arguments-differ
        super(Authenticator, cls).add_parser_arguments(add, default_propagation_seconds=30)
        add('credentials', help='SchlundTech XML Gateway credentials file.')

    def more_info(self):  # pylint: disable=missing-docstring
        return 'This plugin configures a DNS TXT record to respond to a dns-01 challenge using ' + \
               'the SchlundTech XML Gateway API.'

    def _setup_credentials(self):
        self.credentials = self._configure_credentials(
            'credentials',
            'SchlundTech XML Gateway credentials file',
            {
                'user': 'the username for the SchlundTech XML Gateway',
                'password': 'the password for the SchlundTech XML Gateway',
                'context': 'the numeric SchlundTech XML Gateway context to use',
                'token': 'the 2FA token to use, optional',
            }
        )

    def _perform(self, domain, validation_name, validation):
        self._get_gateway_client().add_txt_record(domain, validation_name, validation)

    def _cleanup(self, domain, validation_name, validation):
        self._get_gateway_client().del_txt_record(domain, validation_name, validation)

    def _get_gateway_client(self):
        return _SchlundtechGatewayClient(
            user=self.credentials.conf('user'),
            password=self.credentials.conf('password'),
            context=self.credentials.conf('context'),
            token=self.credentials.conf('token'),
            ttl=self.ttl
        )


class _SchlundtechGatewayClient:
    """
    Encapsulates all communication with the SchlundTech XML Gateway.
    """
    def __init__(self, user, password, context, token=None, ttl=60):
        self.user = user
        self.password = password
        self.context = context
        self.token = token
        self.ttl = ttl
        self._xml = _XML()

    def _auth(self):
        result = {
            'user': self.user,
            'password': self.password,
            'context': self.context,
        }
        if self.token is not None:
            result['token'] = self.token
        return result

    def _call(self, task):
        request = {
            'auth': self._auth(),
            'language': 'en',
            'task': task
        }
        try:
            connection = urlopen(Request(
                url=GATEWAY_URL,
                data=self._xml.tostring('request', request)
            ))
            response = self._xml.fromstring(connection.read())
            if response is not None and response['result'] is not None:
                return response['result']
            else:
                self._log_call_error(request, response, None)
                msg = ('\n' + repr(response)) if response is not None else ''
                raise errors.PluginError(
                    'Unexpected response received from {0}{1}'.format(GATEWAY_URL, msg)
                )
        except HTTPError as e:
            self._log_call_error(request, None, e)
            raise errors.PluginError(
                'Communication error while calling {0}\nResponse: {1}, {2}'.format(GATEWAY_URL, e.code, e.reason)
            )
        except URLError as e:
            self._log_call_error(request, None, e)
            raise errors.PluginError(
                'Communication error while calling {0}\nReason: {1}'.format(GATEWAY_URL, e.reason)
            )

    def _zone_info(self, domain, validation_name):
        result = self._call({
            'code': '0205',
            'zone': {
                'name': self._zone_name(domain, validation_name)
            }
        })
        if result and result['status']['type'] == 'success':
            return result['data']['zone']
        else:
            logger.debug('Failed retrieving zone {0}'.format(domain))
            logger.debug('Response was: ' + repr(result))
            raise errors.PluginError(
                'Unable to find a SchlundTech zone for {0}'.format(domain)
            )

    def add_txt_record(self, domain, validation_name, validation):
        info = self._zone_info(domain, validation_name)
        current_values = self._current_values(info, domain, validation_name)
        if len(current_values) > 0 and validation in current_values:
            logger.debug('{0} already exists with required value'.format(validation_name))
            pass
        else:
            result = self._call({
                'code': '0202001',
                'zone': {
                    'name': self._zone_name(domain, validation_name),
                    'system_ns': info['system_ns']
                },
                'default': {
                    'rr_add': {
                        'name': self._resource_name(domain, validation_name),
                        'type': 'TXT',
                        'value': validation,
                        'ttl': self.ttl
                    },
                    'soa': info['soa']
                }
            })
            if result is None or result['status']['type'] != 'success':
                logger.debug('Failed adding TXT record \'{1}\' for domain \'{0}\''.format(domain, validation_name))
                logger.debug('Response was: ' + repr(result))
                msg = ('\n' + result['status']['text']) if result is not None and 'text' in result['status'] else ''
                raise errors.PluginError(
                    'Unable to add TXT record for {0}: {1}{2}'.format(domain, validation_name, msg)
                )
            else:
                logger.debug('Successfully added TXT record \'{1}\' for domain \'{0}\''.format(domain, validation_name))

    def del_txt_record(self, domain, validation_name, validation):
        info = self._zone_info(domain, validation_name)
        result = self._call({
            'code': '0202001',
            'zone': {
                'name': self._zone_name(domain, validation_name),
                'system_ns': info['system_ns']
            },
            'default': {
                'rr_rem': {
                    'name': self._resource_name(domain, validation_name),
                    'type': 'TXT',
                    'value': validation,
                    'ttl': self.ttl
                }
            }
        })
        if result is None or result['status']['type'] != 'success':
            logger.debug('Failed removing TXT record \'{1}\' for domain \'{0}\''.format(domain, validation_name))
            logger.debug('Response was: ' + repr(result))
            msg = ('\n' + result['status']['text']) if result is not None and 'text' in result['status'] else ''
            raise errors.PluginError(
                'Unable to remove TXT record for {0}: {1}{2}'.format(domain, validation_name, msg)
            )
        else:
            logger.debug('Successfully removed TXT record \'{1}\' for domain \'{0}\''.format(domain, validation_name))

    @staticmethod
    def _zone_name(domain, validation_name):
        return '.'.join(_SchlundtechGatewayClient._fqdn(domain, validation_name)[-2:])

    @staticmethod
    def _resource_name(domain, validation_name):
        return '.'.join(_SchlundtechGatewayClient._fqdn(domain, validation_name)[0:-2])

    @staticmethod
    def _fqdn(domain, validation_name):
        if validation_name.endswith(domain):
            return validation_name.split('.')
        else:
            return validation_name.split('.') + domain.split('.')

    @staticmethod
    def _current_values(info, domain, validation_name):
        name = _SchlundtechGatewayClient._resource_name(domain, validation_name)
        result = []
        if 'rr' in info:
            if type(info['rr']) != list:
                info['rr'] = [info['rr']]  # Convert single values to list
            for rr in info['rr']:
                if rr['name'] == name:
                    result.append(rr['value'])
        return result

    @staticmethod
    def _log_call_error(request, response, error):
        logger.debug('Error calling {0}'.format(GATEWAY_URL))
        logger.debug('Request was:  ' + repr(request['task']))
        if response is not None:
            logger.debug('Response was: ' + repr(response))
        if error is not None:
            logger.debug('Error was:    ' + repr(error))


class _XML:
    """
    Handles XML marshalling and unmarshalling.
    """
    def __init__(self):
        pass

    def tostring(self, tag, obj):
        return Et.tostring(self.serialize(tag, obj))

    def fromstring(self, data):
        return self.deserialize(Et.fromstring(data))

    def _serialize_value(self, e, tag, value):
        t = type(value).__name__ if type(value) is not None else None
        if value is None:
            pass
        elif t in ['str', 'unicode', 'bytes', 'int', 'float']:
            Et.SubElement(e, tag).text = str(value)
        elif t == 'list':
            for item in value:
                self._serialize_value(e, tag, item)
        elif t == 'dict':
            e.append(self.serialize(tag, value))
        elif hasattr(value, '__dict__'):
            e.append(self.serialize(tag, value.__dict__))
        else:
            raise NotImplementedError('Unable to serialize {0}={1} ({2})'.format(tag, value, type(value)))

    def serialize(self, tag, data):
        e = Et.Element(tag)
        for name, value in data.items():
            self._serialize_value(e, name, value)
        return e

    def deserialize(self, e):
        if e.text is not None:
            return e.text
        elif len(e) == 0:
            return None
        else:
            result = {}
            for child in e:
                name = child.tag
                value = self.deserialize(child)
                if name in result:
                    if type(result[name]) == list:
                        result[name].append(value)
                    else:
                        result[name] = [result[name], value]
                else:
                    result[name] = value
            return result



from setuptools import setup
from setuptools import find_packages


version = '0.5.0'

# Remember to update local-oldest-requirements.txt when changing the minimum
# acme/certbot version.
install_requires = [
    'acme>=0.29.0',
    'certbot>=1.0.0',
    'mock',
    'setuptools',
    'zope.interface',
]

docs_extras = [
]

setup(
    name='certbot-dns-schlundtech',
    version=version,
    description="SchlundTech XML Gateway DNS Authenticator plugin for Certbot",
    url='https://github.com/couchtyp/certbot-dns-schlundtech',
    author="Philipp Mueller",
    author_email='pmueller@maildir.eu',
    license='MIT License',
    python_requires='>=2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Plugins',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Security',
        'Topic :: System :: Installation/Setup',
        'Topic :: System :: Networking',
        'Topic :: System :: Systems Administration',
        'Topic :: Utilities',
    ],

    packages=find_packages(),
    include_package_data=True,
    install_requires=install_requires,
    extras_require={
        'docs': docs_extras,
    },
    entry_points={
        'certbot.plugins': [
            'dns-schlundtech = certbot_dns_schlundtech.dns_schlundtech:Authenticator',
            # backward compatability:
            'certbot-dns-schlundtech:auth = certbot_dns_schlundtech.dns_schlundtech:Authenticator',
        ],
    },
    test_suite='certbot_dns_schlundtech',
)
